@@ -121,6 +121,9 @@ routes: Dict[str, str] = {
     "captcha": "/captcha/solve",
     "chapter": "/chapter/{id}",
     "chapter_list": "/chapter",
+    "cover": "/cover/{id}",
+    "cover_list": "/cover",
+    "cover_upload": "/cover/{mangaId}",
     "create_account": "/account/create",
     "finish_recover": "/account/recover/{code}",
     "group": "/group/{id}",
